package net.locusworks.lsproject.object;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;

import net.locusworks.lsproject.object.helper.DijkstraCallback;


/**
 * Keeper and executor of the Dijkstra algorithm.
 */
public class DijkstraEngine {
	public static final int infinity = 100000;
	
	private ArrayList<Router> routers = new ArrayList<Router>();
	private ArrayList<DijkstraCallback> callbacks = new ArrayList<DijkstraCallback>();
	
	/**
	 * Create a comparator to sort which routers take priority.
	 */
	private final Comparator<Router> shortertDistanceComparator = new Comparator<Router>(){
		public int compare(Router firstRouter, Router secondRouter){
			
			int results = getShortestDistance(firstRouter) - getShortestDistance(secondRouter);
			
			return results;
		}
	};

	private final PriorityQueue<Router> unsettledNodes = new PriorityQueue<Router>(
			1, shortertDistanceComparator);

	private final Set<Router> settledNodes = new HashSet<Router>();
	private final Map<Router, Integer> shortestDistances = new HashMap<Router, Integer>();
	private final Map<Router, Router> predecessors = new HashMap<Router, Router>();
	
	/**
	 * Default constructor.
	 */
	public DijkstraEngine() {
	}
	
	/**
	 * Add a callback that will be called when an event happens.
	 * @param dc Callback class to call.
	 */
	public void registerCallback(DijkstraCallback dc) {
		if (!callbacks.contains(dc)) {
			callbacks.add(dc);
		}
	}
	
	/**
	 * Remove a callback from the list.
	 * @param dc Callback to remove.
	 */
	public void unregisterCallback(DijkstraCallback dc) {
		callbacks.remove(dc);
	}
	
	/**
	 * Get the nodes that have their paths determined.
	 * @return Set of settled nodes.
	 */
	public Set<Router> getSettledNodes() {
		return settledNodes;
	}
	
	/**
	 * Get a list of routers which have values associated to them telling what the
	 * shortest distances are from the start router to the key router.
	 * @return Map of router->integers.
	 */
	public Map<Router, Integer> getShortestDistances() {
		return shortestDistances;
	}
	
	/**
	 * Get a list of predecessors.
	 * @return Router->Router mapping.
	 */
	public Map<Router, Router> getPredecessors() {
		return predecessors;
	}
	
	/**
	 * Private method which adds a router and the shortest possible distance from
	 * the start router to the given router.
	 * @param _router Router to add.
	 * @param cost Cost to get to that router.
	 */
	private void setShortestDistance(Router _router, int cost) {
		for (DijkstraCallback dc : callbacks) {
			dc.shortestDistanceUpdated(_router, cost);
		}
		
		unsettledNodes.remove(_router);
		shortestDistances.put(_router, cost);
		unsettledNodes.add(_router);
	}
	 
	/**
	 * 
	 * @param _startRouter
	 */
	private void initialize(Router _startRouter){
		for (DijkstraCallback dc : callbacks) {
			dc.reset();
		}
		
		settledNodes.clear();
        unsettledNodes.clear();
        
        shortestDistances.clear();
        predecessors.clear();
        
        setShortestDistance(_startRouter, 0);
        unsettledNodes.add(_startRouter);
	}
	 
	private boolean isSettled(Router _router){
		return settledNodes.contains(_router);
	}
	 
	private void relaxNeighbors(Router _router){
		for (Router r : Network.getInstance().getNeighboringRouters(_router)){
			if (isSettled(r)) continue;
			if (r == null) continue;
			
			Connection tempConnection = _router.getConnectionByParticipant(_router.getIpAddress(), r.getIpAddress());
			if (tempConnection == null) continue;
			
			int shortDistance = getShortestDistance(_router) + tempConnection.getCost();
			
			if (shortDistance < getShortestDistance(r)){
				setShortestDistance(r, shortDistance);
				setPredecessor(_router, r);
			}
		}
	}
	 
	private void setPredecessor(Router a, Router b){
		for (DijkstraCallback dc : callbacks) {
			dc.predecessorAdded(a, b);
		}
		
        predecessors.put(a, b);
	}
	
	public int getShortestDistance(Router _router){
       Integer d = shortestDistances.get(_router);
       return (d == null) ? infinity : d;
	}
	 
	public Router getPredecessor(Router r){
       return predecessors.get(r);
	}
	
	public void execute(Router startRouter, Router endRouter) {
		initialize(startRouter);
		 
		Router r = null;
		 
		while ((r = unsettledNodes.poll()) != null){
			if (r == endRouter) break;
			 
			settledNodes.add(r);
			for (DijkstraCallback dc : callbacks) {
				dc.settledRouterAdded(r);
			}
			
			relaxNeighbors(r);
		}
	}
	 
	 /**
	  * Find the shortest path between one router and another, returning the path that one must
	  * take to get from the start to the end.
	  * 
	  * Note that this function must run the algorithm multiple times to find how it gets from start to end.
	  * Therefore, in a very large network, this may take a while to execute. On modern computers, this is
	  * not much of a concern. Still, a more efficient algorithm should be used to find start->end along the
	  * quickest path in a more serious system.
	  * @param startRouter Router to start from.
	  * @param endRouter Router to end at.
	  * @return The path (Path object) from the start to the end. If start is not connected to end,
	  * this will return null.
	  */
	public Path findShortestPath(Router startRouter, Router endRouter){
		// Get the array of network routers.
		routers.clear();
		 
		// Run the algorithm once to get the shortest path from start to end..
		execute(startRouter, endRouter);
		 
		if (shortestDistances.get(endRouter) == null) {
			// In this situation, startRouter is not connected to endRouter.
			return null;
		}
		 
		// We now know how much it will cost to get from start to end.
		final int totalCost = shortestDistances.get(endRouter);
		 
		// Leading cost will track how much it costs each step from start to end (additively).
		int leadingCost = 0;
		 
		Router currentRouter = startRouter;
		Router lastRouter = currentRouter;
		 
		// Loop until we reach the last router.
		while (!currentRouter.getIpAddress().equals(endRouter.getIpAddress())) { 
			final ArrayList<Router> neighbors = (ArrayList<Router>)Network.getInstance().getNeighboringRouters(currentRouter);
			// No need to check last router.
			neighbors.remove(lastRouter);
			 
			// Search through the current router's neighbors, attempting to find the shortest path from A to B.
			for (Router r : neighbors) {
				execute(r, endRouter);
				 
				final Connection c = currentRouter.getConnectionByParticipant(
					currentRouter.getIpAddress(), r.getIpAddress());
				 
				// Obtain cost from the current router to the target router.
				final int newCost = shortestDistances.get(endRouter);
				 
				// Test if this router's jump cost is the same as the total cost, taking into
				// consideration how much it cost to get from previous connections to the current router.
				if ((totalCost - newCost - leadingCost) == c.getCost()) {
					// We've found our router -- add it to the path.
					routers.add(Network.getInstance().getRouterByIp(r.getIpAddress()));
					
					leadingCost += c.getCost();
					
					currentRouter = r;
					
					break;
				}
			}
			 
			lastRouter = currentRouter;
		}
		 
		if (!routers.contains(endRouter))
			routers.add(endRouter);
		 
		// Convert the array list to a raw array of routers.
		Router[] list = new Router[routers.size()];
		for (int i = 0; i < routers.size(); i++) {
			list[i] = routers.get(i);
		}
		 
		return new Path(list, totalCost);
	}
}


