package net.locusworks.lsproject.object;
import java.awt.Point;
import java.io.Serializable;
import java.util.ArrayList;

/**
 * A Router is a single entity within the network that is capable of holding one or more connections
 * with various routers across the network.  A router is uniquely identified on the network by it's
 * IP address.
 */
public class Router implements Serializable {
	private static final long serialVersionUID = 8998728841964957750L;
	
	private ArrayList<Connection> connections;
	private String ipAddress;
	private Point position;
	
	private RoutingTable routingTable;
	
	/**
	 * Create a new router.
	 * @param _ipAddress IP address (should be unique) of the router.
	 * @param _position Position of the router on the Network.
	 */
	public Router(String _ipAddress, Point _position) {
		ipAddress = _ipAddress;
		position = _position;
		connections = new ArrayList<Connection>();
		routingTable = new RoutingTable(this);
	}
	
	/**
	 * Return the routing table that this router calcultes.
	 * @return
	 */
	public RoutingTable getRoutingTable() {
		return routingTable;
	}
	
	/**
	 * Get the IP address of this router.
	 * @return String containing the IP address of the router.
	 */
	public String getIpAddress() {
		return ipAddress;
	}
	
	/**
	 * Set a new IP address for this router.
	 * @param _ipAddress New address for the router.
	 */
	public void setIpAddress(String _ipAddress) {
		ipAddress = _ipAddress;
	}
	
	/**
	 * Get the position of the router.
	 * @return Position of the router in a Point object.
	 */
	public Point getPosition() {
		return position;
	}
	
	/**
	 * Set a new position for the router.
	 * @param _position Position to set.
	 */
	public void setPosition(Point _position) {
		position = _position;
	}
	
	/**
	 * Synchronized (thread-safe) method for adding a connection to this router.
	 * @param _newConnection Add a new connection to this router's list of connections.
	 */
	public synchronized void addConnection(Connection _newConnection) {
		connections.add(_newConnection);
		
		// Re-calculate connections.
		routingTable.execute(this, null);
	}
	
	/**
	 * Search through the list of connections and obtain a connection based on the participants.<br>
	 * NOTE: The order does not matter! It will return true regardless of which is the "first" or "second"
	 * participant, assuming such a connection exists.
	 * @param _ipAddress1 One participant's address.
	 * @param _ipAddress2 Another participant's address.
	 * @return Connection object if found, or null if not found.
	 */
	public synchronized Connection getConnectionByParticipant(String _ipAddress1, String _ipAddress2) {
		for (Connection connection : connections) {
			if ((connection .getFirstParticipantAddress() .equals(_ipAddress1)  && 
				 connection .getSecondParticipantAddress().equals(_ipAddress2)) ||
				(connection .getSecondParticipantAddress().equals(_ipAddress1)  &&
				 connection .getFirstParticipantAddress() .equals(_ipAddress2))) {
				// Connection found.
				return connection;
			}
		}
		
		return null;
	}
	
	/**
	 * Get the string-form of Router, which is it's IP address.
	 */
	@Override public String toString() {
		return getIpAddress();
	};
	
	/**
	 * Remove a connection from the list of connections.
	 * @param _ipAddress1 First participant.
	 * @param _ipAddress2 Second participant.
	 */
	public synchronized void removeConnectionByParticipant(String _ipAddress1, String _ipAddress2) {
		connections.remove(getConnectionByParticipant(_ipAddress1, _ipAddress2));
		
		routingTable.execute(this, null);
	}
	
	/**
	 * Get the Router's list of connections.  Useful if the router is deleted and connections
	 * need to be severed.
	 * @return Array list of Connection objects.
	 */
	public synchronized ArrayList<Connection> getConnections() {
		return connections;
	}
	
	/**
	 * Clears the Router's list of connections. 
	 */
	public synchronized void clearConnections(){
		connections.clear();
	}
	
	/**
	 * Simple method for adding connection.
	 * @param router1
	 * @param router2
	 * @param connection
	 */
	public static void addConnection(Router router1, Router router2, Connection connection) {
		router1.addConnection(connection);
		router2.addConnection(connection);
	}
}
